/*****************************************************************************
 * video_output.c : video output thread
 * This module describes the programming interface for video output threads.
 * It includes functions allowing to open a new thread, send pictures to a
 * thread, and destroy a previously oppened video output thread.
 *****************************************************************************
 * Copyright (C) 2000 VideoLAN
 * $Id: video_output.c,v 1.128 2001/05/08 20:38:25 sam Exp $
 *
 * Authors: Vincent Seguin <seguin@via.ecp.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

/*****************************************************************************
 * Preamble
 *****************************************************************************/
#include "defs.h"

#include <errno.h>                                                 /* ENOMEM */
#include <stdlib.h>                                                /* free() */
#include <stdio.h>                                              /* sprintf() */
#include <string.h>                                            /* strerror() */

#ifdef STATS
#   include <sys/times.h>
#endif

#include "config.h"
#include "common.h"
#include "threads.h"
#include "mtime.h"
#include "modules.h"

#include "video.h"
#include "video_output.h"
#include "video_text.h"
#include "video_spu.h"
#include "video_yuv.h"

#include "intf_msg.h"

#include "main.h"

/*****************************************************************************
 * Local prototypes
 *****************************************************************************/
static int      InitThread        ( vout_thread_t *p_vout );
static void     RunThread         ( vout_thread_t *p_vout );
static void     ErrorThread       ( vout_thread_t *p_vout );
static void     EndThread         ( vout_thread_t *p_vout );
static void     DestroyThread     ( vout_thread_t *p_vout, int i_status );

static void     NewPicture        ( picture_t *, int, int, int );

/*****************************************************************************
 * vout_InitBank: initialize the video output bank.
 *****************************************************************************/
void vout_InitBank ( void )
{
    p_vout_bank->i_count = 0;

    vlc_mutex_init( &p_vout_bank->lock );
}

/*****************************************************************************
 * vout_EndBank: empty the video output bank.
 *****************************************************************************
 * This function ends all unused video outputs and empties the bank in
 * case of success.
 *****************************************************************************/
void vout_EndBank ( void )
{
    /* Ask all remaining video outputs to die */
    while( p_vout_bank->i_count )
    {
        vout_DestroyThread(
                p_vout_bank->pp_vout[ --p_vout_bank->i_count ], NULL );
    }

    vlc_mutex_destroy( &p_vout_bank->lock );
}

/*****************************************************************************
 * vout_CreateThread: creates a new video output thread
 *****************************************************************************
 * This function creates a new video output thread, and returns a pointer
 * to its description. On error, it returns NULL.
 * If pi_status is NULL, then the function will block until the thread is ready.
 * If not, it will be updated using one of the THREAD_* constants.
 *****************************************************************************/
vout_thread_t * vout_CreateThread   ( int *pi_status, int i_chroma,
                                      int i_width, int i_height )
{
    vout_thread_t * p_vout;                             /* thread descriptor */
    int             i_status;                               /* thread status */
    int             i_index;                                /* loop variable */

    /* Allocate descriptor */
    p_vout = (vout_thread_t *) malloc( sizeof(vout_thread_t) );
    if( p_vout == NULL )
    {
        intf_ErrMsg( "vout error: vout thread creation returned %s",
                     strerror(ENOMEM) );
        return( NULL );
    }

    /* Choose the best module */
    p_vout->p_module = module_Need( MODULE_CAPABILITY_VOUT, NULL );

    if( p_vout->p_module == NULL )
    {
        intf_ErrMsg( "vout error: no suitable vout module" );
        free( p_vout );
        return( NULL );
    }

#define f p_vout->p_module->p_functions->vout.functions.vout
    p_vout->pf_create     = f.pf_create;
    p_vout->pf_init       = f.pf_init;
    p_vout->pf_end        = f.pf_end;
    p_vout->pf_destroy    = f.pf_destroy;
    p_vout->pf_manage     = f.pf_manage;
    p_vout->pf_display    = f.pf_display;
    p_vout->pf_setpalette = f.pf_setpalette;
#undef f

    /* Initialize thread properties - thread id and locks will be initialized
     * later */
    p_vout->b_die               = 0;
    p_vout->b_error             = 0;
    p_vout->b_active            = 0;
    p_vout->pi_status           = (pi_status != NULL) ? pi_status : &i_status;
    *p_vout->pi_status          = THREAD_CREATE;

    /* Initialize pictures and subpictures - translation tables and functions
     * will be initialized later in InitThread */
    for( i_index = 0; i_index < VOUT_MAX_PICTURES; i_index++)
    {
        p_vout->p_picture[i_index].i_chroma = EMPTY_PICTURE;
        p_vout->p_picture[i_index].i_status = FREE_PICTURE;
    }

    for( i_index = 0; i_index < VOUT_MAX_SUBPICTURES; i_index++)
    {
        p_vout->p_subpicture[i_index].i_type   = EMPTY_SUBPICTURE;
        p_vout->p_subpicture[i_index].i_status = FREE_SUBPICTURE;
    }

    p_vout->i_chroma   = i_chroma;
    p_vout->i_width    = i_width;
    p_vout->i_height   = i_height;
    p_vout->i_pictures = 0;
    p_vout->i_directbuffers = 0;

    /* Create and initialize system-dependant method - this function issues its
     * own error messages */
    if( p_vout->pf_create( p_vout ) )
    {
        module_Unneed( p_vout->p_module );
        free( p_vout );
        return( NULL );
    }

    /* Create thread and set locks */
    vlc_mutex_init( &p_vout->picture_lock );
    vlc_mutex_init( &p_vout->subpicture_lock );
    vlc_mutex_init( &p_vout->change_lock );

    if( vlc_thread_create( &p_vout->thread_id, "video output",
                           (void *) RunThread, (void *) p_vout) )
    {
        intf_ErrMsg("vout error: %s", strerror(ENOMEM));
        p_vout->pf_destroy( p_vout );
        free( p_vout );
        return( NULL );
    }

    /* If status is NULL, wait until the thread is created */
    if( pi_status == NULL )
    {
        do
        {
            msleep( THREAD_SLEEP );
        }while( (i_status != THREAD_READY) && (i_status != THREAD_ERROR)
                && (i_status != THREAD_FATAL) );
        if( i_status != THREAD_READY )
        {
            return( NULL );
        }
    }

    return( p_vout );
}

/*****************************************************************************
 * vout_DestroyThread: destroys a previously created thread
 *****************************************************************************
 * Destroy a terminated thread.
 * The function will request a destruction of the specified thread. If pi_error
 * is NULL, it will return once the thread is destroyed. Else, it will be
 * update using one of the THREAD_* constants.
 *****************************************************************************/
void vout_DestroyThread( vout_thread_t *p_vout, int *pi_status )
{
    int     i_status;                                       /* thread status */

    /* Set status */
    p_vout->pi_status = (pi_status != NULL) ? pi_status : &i_status;
    *p_vout->pi_status = THREAD_DESTROY;

    /* Request thread destruction */
    p_vout->b_die = 1;

    /* If status is NULL, wait until thread has been destroyed */
    if( pi_status == NULL )
    {
        do
        {
            msleep( THREAD_SLEEP );
        }while( (i_status != THREAD_OVER) && (i_status != THREAD_ERROR)
                && (i_status != THREAD_FATAL) );
    }
}

/*****************************************************************************
 * InitThread: initialize video output thread
 *****************************************************************************
 * This function is called from RunThread and performs the second step of the
 * initialization. It returns 0 on success. Note that the thread's flag are not
 * modified inside this function.
 *****************************************************************************/
static int InitThread( vout_thread_t *p_vout )
{
    /* Update status */
    *p_vout->pi_status = THREAD_START;

    vlc_mutex_lock( &p_vout->change_lock );

#ifdef STATS
    p_vout->c_loops = 0;
#endif

    /* Initialize output method, it issues its own error messages */
    if( p_vout->pf_init( p_vout ) )
    {
        return( 1 );
    }

    if( !p_vout->i_directbuffers )
    {
        intf_ErrMsg( "vout error: plugin was unable to allocate at least "
                     "one direct buffer" );
        return( 1 );
    }

    intf_WarnMsg( 1, "vout info: got %i direct buffer(s)",
                  p_vout->i_directbuffers );

    /* Mark thread as running and return */
    p_vout->b_active =          1;
    *p_vout->pi_status =        THREAD_READY;

    intf_DbgMsg("thread ready");
    return( 0 );
}

/*****************************************************************************
 * RunThread: video output thread
 *****************************************************************************
 * Video output thread. This function does only returns when the thread is
 * terminated. It handles the pictures arriving in the video heap and the
 * display device events.
 *****************************************************************************/
static void RunThread( vout_thread_t *p_vout)
{
    int             i_index;                                /* index in heap */
    mtime_t         current_date;                            /* current date */
    mtime_t         display_date;                            /* display date */

    picture_t *     p_picture;                            /* picture pointer */
    boolean_t       b_first_buffer;   /* are we rendering the first buffer ? */

    subpicture_t *  p_subpic;                          /* subpicture pointer */
    subpicture_t *  p_ephemer;        /* youngest ephemer subpicture pointer */
    mtime_t         ephemer_date;                /* earliest subpicture date */

    /*
     * Initialize thread
     */
    p_vout->b_error = InitThread( p_vout );
    if( p_vout->b_error )
    {
        DestroyThread( p_vout, THREAD_ERROR );
        return;
    }

    /*
     * Main loop - it is not executed if an error occured during
     * initialization
     */
    while( (!p_vout->b_die) && (!p_vout->b_error) )
    {
        /* Initialize loop variables */
        p_picture =      NULL;
        p_subpic =       NULL;
        p_ephemer =      NULL;
        ephemer_date =   0;
        display_date =   0;
        current_date =   mdate();
        b_first_buffer = 0;

#ifdef STATS
        p_vout->c_loops++;
        if( !(p_vout->c_loops % VOUT_STATS_NB_LOOPS) )
        {
            intf_Msg( "vout stats: picture heap: %d/%d",
                      p_vout->i_pictures, VOUT_MAX_PICTURES );
        }
#endif

        /*
         * Find the picture to display - this operation does not need lock,
         * since only READY_PICTUREs are handled
         */
        for( i_index = 0; i_index < VOUT_MAX_PICTURES; i_index++ )
        {
            if( (p_vout->p_picture[i_index].i_status == READY_PICTURE) &&
            ( (p_picture == NULL) ||
              (p_vout->p_picture[i_index].date < display_date) ) )
            {
                p_picture = &p_vout->p_picture[i_index];
                display_date = p_picture->date;
            }
        }

        if( p_picture != NULL )
        {
            /* Compute FPS rate */
            p_vout->p_fps_sample[ p_vout->c_fps_samples++ % VOUT_FPS_SAMPLES ]
                = display_date;

            /* XXX: check that we *really* wanted '-' here */
            if( display_date < current_date - p_vout->render_time )
            {
                /* Picture is late: it will be destroyed and the thread
                 * will directly choose the next picture */
                vlc_mutex_lock( &p_vout->picture_lock );
                if( p_picture->i_refcount )
                {
                    /* Pretend we displayed the picture, but don't destroy
                     * it since the decoder might still need it. */
                    p_picture->i_status = DISPLAYED_PICTURE;
                }
                else
                {
                    /* Destroy the picture without displaying it */
                    p_picture->i_status = DESTROYED_PICTURE;
                    p_vout->i_pictures--;
                }
                intf_WarnMsg( 1, "vout warning: late picture skipped (%p)",
                              p_picture );
                vlc_mutex_unlock( &p_vout->picture_lock );

                continue;
            }
            else if( display_date > current_date + VOUT_DISPLAY_DELAY )
            {
                /* A picture is ready to be rendered, but its rendering date
                 * is far from the current one so the thread will perform an
                 * empty loop as if no picture were found. The picture state
                 * is unchanged */
                p_picture    = NULL;
                display_date = 0;
            }
        }

        /*
         * Find the subpictures to display
         */

        /*
         * Perform rendering, sleep and display rendered picture
         */
        if( p_picture != NULL )            /* picture and perhaps subpicture */
        {
            if( p_picture->b_directbuffer )
            {
                if( p_picture->i_refcount )
                {
                    /* Picture is in a direct buffer and is still in use,
                     * we need to copy it to another direct buffer before
                     * displaying it if there are subtitles. */
                    if( p_subpic != NULL )
                    {
                        /* We have subtitles. First copy the picture to
                         * the spare direct buffer, then render the
                         * subtitles. */
                        for( i_index = 0 ;
                             i_index < p_picture->i_planes ;
                             i_index++ )
                        {
                            memcpy( p_vout->p_picture[0].planes[
                                                       i_index ].p_data,
                                    p_picture->planes[ i_index ].p_data,
                                    p_picture->planes[ i_index ].i_bytes );
                        }

                        /* XXX: subtitles */

                        b_first_buffer = 1;
                    }
                    else
                    {
                        /* No subtitles, picture is in a directbuffer so
                         * we can display it directly even if it is still
                         * in use. */
                    }
                }
                else
                {
                    /* Picture is in a direct buffer but isn't used by the
                     * decoder. We can safely render subtitles on it and
                     * display it. */

                    /* XXX: subtitles */
                }
            }
            else
            {
                if( ( p_picture->i_chroma == p_vout->p_picture[0].i_chroma ) &&
                    ( p_picture->i_height == p_vout->p_picture[0].i_height ) &&
                    ( p_picture->i_width  == p_vout->p_picture[0].i_width ) )
                {
                    /* Picture is not in a direct buffer, but is exactly the
                     * same size as the direct buffers. A memcpy() is enough,
                     * then render the subtitles. */
                    for( i_index = 0 ;
                         i_index < p_picture->i_planes ;
                         i_index++ )
                    {
                        memcpy( p_vout->p_picture[0].planes[ i_index ].p_data,
                                p_picture->planes[ i_index ].p_data,
                                p_picture->planes[ i_index ].i_bytes );
                    }

                    /* XXX: subtitles */

                    b_first_buffer = 1;
                }
                else
                {
                    /* Worst case: picture is not in a direct buffer, and
                     * needs to be converted to another size/chroma. Then the
                     * subtitles need to be rendered as well. */

                    /* XXX: render to direct buffer */

                    /* XXX: subtitles */

                    b_first_buffer = 1;
                }
            }
        }

        /*
         * Sleep, wake up and display rendered picture
         */
        if( display_date != 0 )
        {
            /* Store render time using Bresenham algorithm */
            p_vout->render_time += mdate() - current_date;
            p_vout->render_time >>= 1;
        }

        /* Give back change lock */
        vlc_mutex_unlock( &p_vout->change_lock );

        /* Sleep a while or until a given date */
        if( display_date != 0 )
        {
            mwait( display_date - VOUT_MWAIT_TOLERANCE );
        }
        else
        {
            msleep( VOUT_IDLE_SLEEP );
        }

        /* On awakening, take back lock and send immediately picture
         * to display. */
        vlc_mutex_lock( &p_vout->change_lock );

#ifdef TRACE_VOUT
        intf_DbgMsg( "picture %p, subpicture %p", p_picture, p_subpic );
#endif

        if( p_picture != NULL )
        {
            /* Display the previously rendered buffer. */
            if( b_first_buffer )
            {
                p_vout->pf_display( p_vout, &p_vout->p_picture[0] );
            }
            else
            {
                if( p_picture->b_directbuffer )
                {
                    p_vout->pf_display( p_vout, p_picture );
                }
                else
                {
                    printf( "vout error: trying to display memory\n" );
                    //intf_ErrMsg( "vout error: trying to display memory" );
                }
            }

            /* Remove picture from heap */
            vlc_mutex_lock( &p_vout->picture_lock );
            if( p_picture->i_refcount )
            {
                p_picture->i_status = DISPLAYED_PICTURE;
            }
            else
            {
                p_picture->i_status = DESTROYED_PICTURE;
                p_vout->i_pictures--;
            }
            vlc_mutex_unlock( &p_vout->picture_lock );
        }

        /*
         * Check events and manage thread
         */
        if( p_vout->pf_manage( p_vout ) )
        {
            /* A fatal error occured, and the thread must terminate immediately,
             * without displaying anything - setting b_error to 1 causes the
             * immediate end of the main while() loop. */
            p_vout->b_error = 1;
        }
    }

    /*
     * Error loop - wait until the thread destruction is requested
     */
    if( p_vout->b_error )
    {
        ErrorThread( p_vout );
    }

    /* End of thread */
    EndThread( p_vout );
    DestroyThread( p_vout, THREAD_OVER );
    intf_DbgMsg( "thread end" );
}

/*****************************************************************************
 * ErrorThread: RunThread() error loop
 *****************************************************************************
 * This function is called when an error occured during thread main's loop. The
 * thread can still receive feed, but must be ready to terminate as soon as
 * possible.
 *****************************************************************************/
static void ErrorThread( vout_thread_t *p_vout )
{
    /* Wait until a `die' order */
    while( !p_vout->b_die )
    {
        /* Sleep a while */
        msleep( VOUT_IDLE_SLEEP );
    }
}

/*****************************************************************************
 * EndThread: thread destruction
 *****************************************************************************
 * This function is called when the thread ends after a sucessful
 * initialization. It frees all ressources allocated by InitThread.
 *****************************************************************************/
static void EndThread( vout_thread_t *p_vout )
{
    int     i_index;                                        /* index in heap */

    /* Store status */
    *p_vout->pi_status = THREAD_END;

#ifdef STATS
    {
        struct tms cpu_usage;
        times( &cpu_usage );

        intf_Msg( "vout stats: cpu usage (user: %d, system: %d)",
                  cpu_usage.tms_utime, cpu_usage.tms_stime );
    }
#endif

    /* Destroy all remaining pictures */
    for( i_index = p_vout->i_directbuffers;
         i_index < VOUT_MAX_PICTURES;
         i_index++ )
    {
        if( p_vout->p_picture[i_index].i_status != FREE_PICTURE )
        {
            free( p_vout->p_picture[i_index].planes[0].p_data );
        }
    }

    /* Destroy all remaining subpictures */
    for( i_index = 0; i_index < VOUT_MAX_SUBPICTURES; i_index++ )
    {
        if( p_vout->p_subpicture[i_index].i_status != FREE_SUBPICTURE )
        {
            free( p_vout->p_subpicture[i_index].p_data );
        }
    }

    /* Destroy translation tables */
    p_vout->pf_end( p_vout );

    /* Release the change lock */
    vlc_mutex_unlock( &p_vout->change_lock );
}

/*****************************************************************************
 * DestroyThread: thread destruction
 *****************************************************************************
 * This function is called when the thread ends. It frees all ressources
 * allocated by CreateThread. Status is available at this stage.
 *****************************************************************************/
static void DestroyThread( vout_thread_t *p_vout, int i_status )
{
    int *pi_status;                                         /* status adress */

    /* Store status adress */
    pi_status = p_vout->pi_status;

    /* Destroy thread structures allocated by Create and InitThread */
    p_vout->pf_destroy( p_vout );

    /* Destroy the locks */
    vlc_mutex_destroy( &p_vout->picture_lock );
    vlc_mutex_destroy( &p_vout->subpicture_lock );
    vlc_mutex_destroy( &p_vout->change_lock );

    /* Release the module */
    module_Unneed( p_vout->p_module );

    /* Free structure */
    free( p_vout );
    *pi_status = i_status;
}

/*****************************************************************************
 * vout_DisplayPicture: display a picture
 *****************************************************************************
 * Remove the reservation flag of a picture, which will cause it to be ready for
 * display. The picture won't be displayed until vout_DatePicture has been
 * called.
 *****************************************************************************/
void vout_DisplayPicture( vout_thread_t *p_vout, picture_t *p_picture )
{
    vlc_mutex_lock( &p_vout->picture_lock );
    switch( p_picture->i_status )
    {
    case RESERVED_PICTURE:
        p_picture->i_status = RESERVED_DISP_PICTURE;
        break;
    case RESERVED_DATED_PICTURE:
        p_picture->i_status = READY_PICTURE;
        break;
#ifdef DEBUG
    default:
        intf_ErrMsg("error: picture %p has invalid status %d", p_picture, p_picture->i_status );
        break;
#endif
    }

#ifdef TRACE_VOUT
    intf_DbgMsg("picture %p", p_picture);
#endif
    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_DatePicture: date a picture
 *****************************************************************************
 * Remove the reservation flag of a picture, which will cause it to be ready
 * for display. The picture won't be displayed until vout_DisplayPicture has
 * been called.
 *****************************************************************************/
void vout_DatePicture( vout_thread_t *p_vout,
                       picture_t *p_picture, mtime_t date )
{
#ifdef TRACE_VOUT
    char        psz_date[ MSTRTIME_MAX_SIZE ];                       /* date */
#endif

    vlc_mutex_lock( &p_vout->picture_lock );
    p_picture->date = date;
    switch( p_picture->i_status )
    {
    case RESERVED_PICTURE:
        p_picture->i_status = RESERVED_DATED_PICTURE;
        break;
    case RESERVED_DISP_PICTURE:
        p_picture->i_status = READY_PICTURE;
        break;
#ifdef DEBUG
    default:
        intf_ErrMsg("error: picture %p has invalid status %d", p_picture, p_picture->i_status );
        break;
#endif
    }

#ifdef TRACE_VOUT
    intf_DbgMsg("picture %p, display date: %s", p_picture, mstrtime( psz_date, p_picture->date) );
#endif
    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_CreatePicture: allocate a picture in the video output heap.
 *****************************************************************************
 * This function create a reserved image in the video output heap.
 * A null pointer is returned if the function fails. This method provides an
 * already allocated zone of memory in the picture data fields. It needs locking
 * since several pictures can be created by several producers threads.
 *****************************************************************************/
picture_t *vout_CreatePicture( vout_thread_t *p_vout, int i_chroma,
                               int i_width, int i_height )
{
    int         i_picture;                                  /* picture index */
    picture_t * p_picture;
    picture_t * p_free_picture = NULL;                 /* first free picture */
    picture_t * p_destroyed_picture = NULL;       /* first destroyed picture */

    /* Get lock */
    vlc_mutex_lock( &p_vout->picture_lock );

    /*
     * Look for an empty place. XXX: we start at 1 because the first
     * directbuffer is reserved for memcpy()ed pictures.
     */
    for( i_picture = 1; i_picture < VOUT_MAX_PICTURES; i_picture++ )
    {
        p_picture = p_vout->p_picture + i_picture;

        if( p_picture->i_status == DESTROYED_PICTURE )
        {
            /* Picture is marked for destruction, but is still allocated.
             * Note that if width and type are the same for two pictures,
             * chroma_width should also be the same */
            if( ( p_picture->i_chroma == i_chroma ) &&
                ( p_picture->i_height == i_height ) &&
                ( p_picture->i_width  == i_width ) )
            {
                /* Memory size do match : memory will not be reallocated,
                 * and function can end immediately - this is the best
                 * possible case, since no memory allocation needs to be
                 * done */
                p_picture->i_status = RESERVED_PICTURE;
                p_vout->i_pictures++;

                vlc_mutex_unlock( &p_vout->picture_lock );

                return( p_picture );
            }
            else if( ( p_destroyed_picture == NULL )
                     && !p_picture->b_directbuffer )
            {
                /* Memory size do not match, but picture index will be kept in
                 * case no other place are left */
                p_destroyed_picture = p_picture;
            }
        }
        else if( ( p_free_picture == NULL )
                 && ( p_picture->i_status == FREE_PICTURE ) )
        {
            /* Picture is empty and ready for allocation */
            p_free_picture = p_picture;
        }
    }

    /* If no free picture is available, use a destroyed picture */
    if( ( p_free_picture == NULL ) && ( p_destroyed_picture != NULL ) )
    {
        /* No free picture or matching destroyed picture has been found, but
         * a destroyed picture is still avalaible */
        free( p_destroyed_picture->planes[0].p_data );
        p_destroyed_picture->i_planes = 0;
        p_free_picture = p_destroyed_picture;
    }

    /*
     * Prepare picture
     */
    if( p_free_picture != NULL )
    {
        NewPicture( p_free_picture, i_chroma, i_width, i_height );

        if( p_free_picture->i_planes != 0 )
        {
            /* Copy picture information, set some default values */
            p_free_picture->i_chroma              = i_chroma;
            p_free_picture->i_width               = i_width;
            p_free_picture->i_height              = i_height;
            p_free_picture->i_status              = RESERVED_PICTURE;
            p_free_picture->i_matrix_coefficients = 1;
            p_free_picture->i_aspect_ratio        = AR_SQUARE_PICTURE;
            p_free_picture->i_refcount            = 0;
            p_vout->i_pictures++;
        }
        else
        {
            /* Memory allocation failed : set picture as empty */
            p_free_picture->i_chroma =  EMPTY_PICTURE;
            p_free_picture->i_status =  FREE_PICTURE;
            p_free_picture =            NULL;

            intf_ErrMsg( "vout error: picture allocation failed" );
        }

        vlc_mutex_unlock( &p_vout->picture_lock );

        /* Initialize mutex */
        vlc_mutex_init( &(p_free_picture->lock_deccount) );

        return( p_free_picture );
    }

    /* No free or destroyed picture could be found, but the decoder
     * will try again in a while. */
    vlc_mutex_unlock( &p_vout->picture_lock );

    return( NULL );
}

/*****************************************************************************
 * vout_DestroyPicture: remove a permanent or reserved picture from the heap
 *****************************************************************************
 * This function frees a previously reserved picture or a permanent
 * picture. It is meant to be used when the construction of a picture aborted.
 * Note that the picture will be destroyed even if it is linked !
 *****************************************************************************/
void vout_DestroyPicture( vout_thread_t *p_vout, picture_t *p_picture )
{
    vlc_mutex_lock( &p_vout->picture_lock );

#ifdef DEBUG
    /* Check if picture status is valid */
    if( (p_picture->i_status != RESERVED_PICTURE) &&
        (p_picture->i_status != RESERVED_DATED_PICTURE) &&
        (p_picture->i_status != RESERVED_DISP_PICTURE) )
    {
        intf_ErrMsg( "error: picture %p has invalid status %d",
                     p_picture, p_picture->i_status );
    }
#endif

    p_picture->i_status = DESTROYED_PICTURE;
    p_vout->i_pictures--;

    /* destroy the lock that had been initialized in CreatePicture */
    vlc_mutex_destroy( &(p_picture->lock_deccount) );

    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_LinkPicture: increment reference counter of a picture
 *****************************************************************************
 * This function increments the reference counter of a picture in the video
 * heap. It needs a lock since several producer threads can access the picture.
 *****************************************************************************/
void vout_LinkPicture( vout_thread_t *p_vout, picture_t *p_picture )
{
    vlc_mutex_lock( &p_vout->picture_lock );
    p_picture->i_refcount++;

#ifdef TRACE_VOUT
    intf_DbgMsg( "picture %p refcount=%d", p_picture, p_picture->i_refcount );
#endif

    vlc_mutex_unlock( &p_vout->picture_lock );
}

/*****************************************************************************
 * vout_UnlinkPicture: decrement reference counter of a picture
 *****************************************************************************
 * This function decrement the reference counter of a picture in the video heap.
 *****************************************************************************/
void vout_UnlinkPicture( vout_thread_t *p_vout, picture_t *p_picture )
{
    vlc_mutex_lock( &p_vout->picture_lock );
    p_picture->i_refcount--;

#ifdef TRACE_VOUT
    if( p_picture->i_refcount < 0 )
    {
        intf_DbgMsg( "error: refcount < 0" );
        p_picture->i_refcount = 0;
    }
#endif

    if( ( p_picture->i_refcount == 0 ) &&
        ( p_picture->i_status == DISPLAYED_PICTURE ) )
    {
        p_picture->i_status = DESTROYED_PICTURE;
        p_vout->i_pictures--;
    }

#ifdef TRACE_VOUT
    intf_DbgMsg( "picture %p refcount=%d", p_picture, p_picture->i_refcount );
#endif

    vlc_mutex_unlock( &p_vout->picture_lock );
}

/* Following functions are local */

/*****************************************************************************
 * NewPicture: allocate a new picture in the heap.
 *****************************************************************************
 * This function allocates a fake direct buffer in memory, which can be
 * used exactly like a video buffer. The video output thread then manages
 * how it gets displayed.
 *****************************************************************************/
static void NewPicture( picture_t *p_picture, int i_chroma,
                        int i_width, int i_height )
{
    int i_data_size = 0;

    p_picture->i_size = i_width * i_height;

    /* Calculate coordinates */
    switch( i_chroma )
    {
        case YUV_420_PICTURE:        /* YUV 420: 1,1/4,1/4 samples per pixel */
            p_picture->i_chroma_size = p_picture->i_size / 4;
            p_picture->i_chroma_width = i_width / 2;
            break;

        case YUV_422_PICTURE:        /* YUV 422: 1,1/2,1/2 samples per pixel */
            p_picture->i_chroma_size = p_picture->i_size / 2;
            p_picture->i_chroma_width = i_width / 2;
            break;

        case YUV_444_PICTURE:            /* YUV 444: 1,1,1 samples per pixel */
            p_picture->i_chroma_size = p_picture->i_size;
            p_picture->i_chroma_width = i_width;
            break;

        default:
            intf_ErrMsg("error: unknown chroma type %d", i_chroma );
            p_picture->i_planes = 0;
            return;
    }

    /* Allocate memory */
    switch( i_chroma )
    {
        case YUV_420_PICTURE:        /* YUV 420: 1,1/4,1/4 samples per pixel */
        case YUV_422_PICTURE:        /* YUV 422: 1,1/2,1/2 samples per pixel */
        case YUV_444_PICTURE:            /* YUV 444: 1,1,1 samples per pixel */

            i_data_size = p_picture->i_size + 2 * p_picture->i_chroma_size;

            /* The Y plane */
            p_picture->planes[ Y_PLANE ].i_bytes =
                 p_picture->i_size * sizeof(pixel_data_t);
            p_picture->planes[ Y_PLANE ].p_data =
                 malloc( i_data_size * sizeof(pixel_data_t) * 4 );
            /* The U plane */
            p_picture->planes[ U_PLANE ].i_bytes =
                 p_picture->i_chroma_size * sizeof(pixel_data_t);
            p_picture->planes[ U_PLANE ].p_data =
                 p_picture->planes[ Y_PLANE ].p_data + p_picture->i_size;
            /* The V plane */
            p_picture->planes[ V_PLANE ].i_bytes =
                 p_picture->i_chroma_size * sizeof(pixel_data_t);
            p_picture->planes[ V_PLANE ].p_data =
                 p_picture->planes[ U_PLANE ].p_data + p_picture->i_chroma_size;

            p_picture->i_planes = 3;

            break;

        default:
            p_picture->i_planes = 0;

            break;
    }
}

